/*
** Module   : DBUtils.java
** Abstract : Static helper methods for DB services
**
** Copyright (c) 2004-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060712   @28043 Created initial version. Static utility methods for the persistence
**                           framework.
** 002 ECF 20060714   @28136 Made class and most methods public. Access is required from the
**                           schema package for import.
** 003 ECF 20070416   @33031 Added handleInterrupt(). Inspects an exception and throws
**                           StopConditionException if it finds InterruptedException within the
**                           exception's causal chain.
** 004 EVL 20070609   @34005 Adding explicit import of the class org.hibernate.type.Type to
**                           eliminate conflict with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 005 ECF 20070705   @34358 Enhanced makeTypeArray(). Entity parameters are now allowed.
** 006 ECF 20071207   @36260 Changed findDMOGetterMethod() signature. Now
**                           throws PersistenceException rather than IllegalArgumentException.
** 007 ECF 20080310   @37469 Added new database dialect support. Method composePropertyName()
**                           now supports computed columns. Also added new variant of method
**                           getDMOPropertyType().
** 008 ECF 20080318   @37542 Fixed makeTypeArray(). It was not properly handling unresolved
**                           (i.e. Resolvable) query substitution parameters.
** 009 ECF 20080330   @37731 Fixed composePropertyName() (package private variant). It was
**                           qualifying property name with an alias at the wrong point.
** 010 ECF 20080510   @38483 Added new methods. entityForDMOInterface() and dmoClassForEntity().
** 011 ECF 20080806   @39318 Added rowid support. New mappings for 64-bit integers. Import
**                           persist.type package, since UserType classes moved.
** 012 CA  20080815   @39454 Changed handleInterrupt to handleException.
**                           The method will handle database connection errors beside interrupt.
** 013 GES 20090424   @41936 Import change.
** 014 SVL 20090225   @41365 Into TYPES map added mapping for recid.
** 015 ECF 20090525   @42972 Added composePropertyName() variant. Added dirty parameter to
**                           differentiate between primary and dirty database dialects when
**                           composing property names.
** 016 ECF 20090717   @43228 Created workaround for remote dirty database lookup failure.
**                           We drop back to using primary database's dialect in
**                           composePropertyName().
** 017 ECF 20090728   @43416 Fixed getDMOPropertyType(). RecordBuffer change
**                           had regressed this method.
** 018 ECF 20090817   @43682 Added new *Field types to TYPES map.
** 019 VMN 20130416          Added method getGetterSetterType().
** 020 SVL 20130331          Upgraded to Hibernate 4.
** 021 OM  20130430          Added new int64 and Int64Field to TYPES map.
** 022 OM  20130527          Remapped DatetimeTzUserType as CompositeCustomType.
**                           For the moment this is disabled (1584#150).
** 023 SVL 20130716          Use system class loader in dmoClassForEntity.
** 024 ECF 20131010          Added metadata support.
** 025 OM  20140410          Fixed support case-sensitive fields in indexes.
** 026 VMN 20140502          Added enhance schema name conversion support for hint "escape"
**                           attribute. Added helper methods for processing backticks.
** 027 OM  20140513          Added DEFAULT_MAX_INDEX_SIZE & TABLE_ID_FIELD_SIZE constant fields.
** 028 OM  20140623          Added new parameter to composePropertyName for insering the proper
**                           rtrim sql function based of the dialect used.
** 029 VMN 20141127          Renamed DatetimeTzField.class to DatetimetzField.class.
** 030 ECF 20150220          Replaced Apache commons logging with J2SE logging.
** 031 SVL 20150503          Added getFieldForAddLikeColumn functions.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.cfg.*;
import org.hibernate.type.*;
import org.hibernate.type.Type;
import com.goldencode.util.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.type.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;

/**
 * A collection useful static methods used to support the persistence
 * framework.
 */
public final class DBUtils
{
   /**
    * The default maximum permitted size for all combined fields of an index. By default this
    * is equal to Progress version 10.x limit of 1971 bytes (1972 is already in the red zone).
    */
   public static final int DEFAULT_MAX_INDEX_SIZE = 1971;
   
   /**
    * The size of the id primary synthetic field (DatabaseManager.PRIMARY_KEY) added to tables,
    * used by non-unique indexes. This is used for computing the index-key. Because it's of
    * BIGINT / INT64 type so the it will always take 8 bytes.
    */
   public static final int TABLE_ID_FIELD_SIZE = 8;
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(DBUtils.class.getName());
   
   /** Mapping of wrapper data types to Hibernate <code>Type</code>s */
   private static final Map<Class<?>, Type> TYPES = new HashMap<>();
   
   // Initialize the TYPES map.
   static
   {
      TYPES.put(integer.class   , new CustomType(new IntegerUserType()));
      TYPES.put(recid.class     , new CustomType(new IntegerUserType()));
      TYPES.put(decimal.class   , new CustomType(new DecimalUserType()));
      TYPES.put(character.class , new CustomType(new CharacterUserType()));
      TYPES.put(logical.class   , new CustomType(new LogicalUserType()));
      TYPES.put(date.class      , new CustomType(new DateUserType()));
      TYPES.put(raw.class       , new CustomType(new RawUserType()));
      TYPES.put(rowid.class     , new CustomType(new RowIDUserType()));
      TYPES.put(int64.class     , new CustomType(new Int64UserType()));
      TYPES.put(datetime.class  , new CustomType(new DatetimeUserType()));
      // TYPES.put(datetimetz.class, new CompositeCustomType(new DatetimeTzUserType()));
      TYPES.put(datetimetz.class, new CustomType(new DatetimeUserType()));
      TYPES.put(Integer.class   , StandardBasicTypes.INTEGER);
      TYPES.put(Long.class      , StandardBasicTypes.LONG);
      TYPES.put(IntegerField.class   , new CustomType(new IntegerUserType()));
      TYPES.put(DecimalField.class   , new CustomType(new DecimalUserType()));
      TYPES.put(CharacterField.class , new CustomType(new CharacterUserType()));
      TYPES.put(LogicalField.class   , new CustomType(new LogicalUserType()));
      TYPES.put(DateField.class      , new CustomType(new DateUserType()));
      TYPES.put(RawField.class       , new CustomType(new RawUserType()));
      TYPES.put(Int64Field.class     , new CustomType(new Int64UserType()));
      TYPES.put(DatetimeField.class  , new CustomType(new DatetimeUserType()));
      //TYPES.put(DatetimeTzField.class, new CompositeCustomType(new DatetimeTzUserType()));
      TYPES.put(DatetimetzField.class, new CustomType(new DatetimeUserType()));
   }
   
   /**
    * Get the mapping of Java class to Hibernate data type used by P2J.
    * 
    * @return  Unmodifiable map of classes to Hibernate types.
    */
   public static Map<Class<?>, Type> getHibernateTypeMap()
   {
      return Collections.unmodifiableMap(TYPES);
   }
   
   /**
    * Convenience method to create an array of Hibernate <code>Type</code>
    * objects to match the given array of substitution parameters.  It is
    * assumed that neither <code>args</code> nor any of its elements are
    * <code>null</code>.
    *
    * @param   args
    *          Array of HQL query substitution parameters.
    *
    * @return  Array of matching, Hibernate <code>Type</code> instances.
    * 
    * @throws  IllegalArgumentException
    *          if <code>args</code> contains a parameter of unknown type.
    */
   public static Type[] makeTypeArray(Object[] args)
   {
      return makeTypeArray(args, new Type[args.length]);
   }
   
   /**
    * Convenience method to populate an array of Hibernate <code>Type</code>
    * objects to match the given array of substitution parameters.  It is
    * assumed that neither <code>args</code> nor any of its elements are
    * <code>null</code>.
    * <p>
    * This implementation handles both resolved (<code>BaseDataType</code>)
    * and unresolved (<code>Resolvable</code>) arguments.  The latter is
    * possible because we only care about the data type, not the actual value,
    * of the parameters.
    * <p>
    * If any element in the <code>types</code> array is non-<code>null</code>,
    * it is skipped.
    *
    * @param   args
    *          Array of HQL query substitution parameters.
    * @param   types
    *          Array of the same size as <code>args</code>;  some elements
    *          may be pre-populated.
    *
    * @return  Array of matching, Hibernate <code>Type</code> instances.
    * 
    * @throws  IllegalArgumentException
    *          if <code>args</code> contains a parameter of unknown type.
    */
   public static Type[] makeTypeArray(Object[] args, Type[] types)
   {
      int len = args.length;
      for (int i = 0; i < len; i++)
      {
         if (types[i] != null)
         {
            // Type already known;  don't replace it.
            continue;
         }
         
         Object next = args[i];
         Class<?> cls = (next instanceof Resolvable
                         ? ((Resolvable) next).getType()
                         : next.getClass());
         
         // Look up the Hibernate type associated with the Java type.
         Type type = TYPES.get(cls);
         
         // If no type matched, assume we are using an entity parameter.
         if (type == null)
         {
            // This will throw IllegalArgumentException if class cannot be
            // found among mapped, persistent classes.
            Class normalClass = DatabaseManager.normalizeDMOClass(cls);
            String schema = DatabaseManager.getSchemaByClass(normalClass);
            Configuration cfg = DatabaseManager.getConfigBySchema(schema);
            type = cfg.getTypeResolver().getTypeFactory().manyToOne(normalClass.getName());
         }

         types[i] = type;
      }
      
      return types;
   }
   
   /**
    * Report the data type of a DMO property. Reflection is used to look up
    * the field which backs the specified property. This will not work for
    * list/composite properties, since these are not backed by a simple field.
    * In these cases, the indexed getter method is introspected and the data
    * type which it returns is reported.
    * <p>
    * Note:  this method will only work while the {@link RecordBuffer}
    * represented by the variable name <code>dmoAlias</code> has had a scope
    * opened upon it by business logic, within the current context.
    *
    * @param   dmoAlias
    *          Variable
    * @param   name
    *          Unqualified DMO property name.
    *
    * @return  Class which indicates the data type of the specified property.
    *
    * @throws  NullPointerException
    *          if <code>dmoAlias</code> does not represent a record buffer
    *          whose scope is currently open.
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static Class<?> getDMOPropertyType(String dmoAlias, String name)
   throws PersistenceException
   {
      RecordBuffer buffer = BufferManager.get().lookupBuffer(dmoAlias);
      if (buffer == null)
      {
         throw new NullPointerException(
            "No record buffer found for alias '" + dmoAlias + "'");
      }
      buffer.initialize();
      
      return getDMOPropertyType(buffer.getDMOImplementationClass(), name);
   }
   
   /**
    * Report the data type of a DMO property. Reflection is used to look up
    * the field which backs the specified property. This will not work for
    * list/composite properties, since these are not backed by a simple field.
    * In these cases, the indexed getter method is introspected and the data
    * type which it returns is reported.
    *
    * @param   dmoClass
    *          DMO implementation class.
    * @param   name
    *          Unqualified DMO property name.
    *
    * @return  Class which indicates the data type of the specified property.
    *
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static Class<?> getDMOPropertyType(Class<?> dmoClass, String name)
   throws PersistenceException
   {
      try
      {
         Class<?> propClass = null;
         
         // The easy way (by field)...
         try
         {
            propClass = dmoClass.getDeclaredField(name).getType();
         }
         catch (NoSuchFieldException nsfExc)
         {
         }
         
         // ...or the hard way (by indexed getter method).
         if (propClass == null || List.class.isAssignableFrom(propClass))
         {
            Class<?>[] parmTypes = new Class<?>[] { Integer.TYPE };
            Method getter = findDMOGetterMethod(dmoClass, name, parmTypes);
            propClass = getter.getReturnType();
         }
         
         return propClass;
      }
      catch (Exception exc)
      {
         String msg = "Error detecting DMO field type:  " + dmoClass.getName() + "." + name;
                      
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, msg, exc);
         }
         
         throw new PersistenceException(msg, exc);
      }
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name is enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is emitted.
    *
    * @param   isCharacter
    *          <code>true</code> for text properties;  <code>false</code> for
    *          non-text properties.
    * @param   ignoreCase
    *          <code>true</code> if the name expression must be wrapped in an
    *          uppercasing function;  <code>false</code> for case-sensitive
    *          text properties, as well as for non-text properties.
    * @param   computedColumnPrefix
    *          If a computed column is needed instead of embedding the DMO
    *          property name within a function, this parameter should be the
    *          non-null prefix to be prepended before the property name.
    *          If <code>null</code>, it is assumed function use is acceptable.
    * @param   alias
    *          Optional alias used to qualify <code>name</code>.  Ignored if <code>null</code>.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    * @param   dialect
    *          The dialect used for <code>rtrim</code> sql function. If <code>isCharacter</code>
    *          is <code>false</code> this parameter is ignored.
    */
   public static void composePropertyName(boolean isCharacter,
                                          boolean ignoreCase,
                                          String computedColumnPrefix,
                                          String alias,
                                          String name,
                                          StringBuilder buf,
                                          P2JDialect dialect)
   {
      String qualifier = "";
      if (alias != null)
      {
         qualifier = alias + '.';
      }
      
      if (!isCharacter)
      {
         buf.append(qualifier).append(name);
         
         return;
      }
      
      if (computedColumnPrefix != null)
      {
         buf.append(qualifier);
         buf.append(getPrefixedParameter(name, computedColumnPrefix));
         
         return;
      }
      
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      
      // use dialect-specific rtrim function
      dialect.addRtrimmedExpression(qualifier + name, buf);
      
      if (ignoreCase)
      {
         buf.append(")");
      }
   }
   
   /**
    * Get the DMO implementation class for the given DMO entity name.
    * 
    * @param   entityName
    *          DMO entity name.
    * 
    * @return  DMO implementation class.
    * 
    * @throws  IllegalArgumentException
    *          if <code>entityName</code> does not represent a loaded class,
    *          or if there is any other error loading the class.
    */
   public static Class<?> dmoClassForEntity(String entityName)
   {
      try
      {
         return Class.forName(entityName, true, ClassLoader.getSystemClassLoader());
      }
      catch (Exception exc)
      {
         throw new IllegalArgumentException("Invalid DMO entity name:  " + entityName, exc);
      }
   }
   
   /**
    * Generate the DMO implementation class name (the entity name), given a
    * DMO interface.  This codifies the convention of adding the text "Impl"
    * to the end of the interface name to come up with the implementation
    * name.
    * 
    * @param   dmoIface
    *          DMO interface.
    * 
    * @return  DMO implementation class name.
    */
   public static String entityForDMOInterface(Class<?> dmoIface)
   {
      return dmoIface.getName() + "Impl";
   }

   /**
    * Extracts index name from index name optionally containing prefix "idx__" and backticks.
    *
    * @param    indexName
    *           Index name.
    *
    * @return   Extracts index name from index name optionally containing prefix "idx__" and
    *           backticks.
    */
   public static String extractIndexName(String indexName)
   {
      StringBuilder buf = new StringBuilder();
      boolean escaped = indexName.startsWith("`");
      if (escaped)
      {
         indexName = indexName.substring(1, indexName.length() - 1);
      }

      if (indexName.startsWith("idx__"))
      {
         // eliminate the prefix
         indexName = indexName.substring("idx__".length());
      }

      buf.append(indexName);
      if (escaped)
      {
         buf.insert('`', 0);
         buf.append('`');
      }
      return buf.toString();
   }

   /**
    * Returns a formatted string using the specified format string and parameter optionally
    * containing backticks.
    *
    * @param    format
    *           Format string.
    * @param    parameter
    *           String parameter optionally containing backticks..
    *
    * @return   Returns a formatted string using the specified format string and parameter
    *           optionally containing backticks.
    */
   public static String formatWithEscapedParameter(String format, String parameter)
   {
      return formatWithEscapedParameter(format, parameter, null);
   }

   /**
    * Returns a formatted string using the specified format string and arguments optionally
    * containing backticks.
    *
    * @param    format
    *           Format string.
    * @param    parameter
    *           String parameter optionally containing backticks..
    * @param    index
    *           Nullable value.
    *
    * @return   Returns a formatted string using the specified format string and arguments
    *           optionally containing backticks.
    */
   public static String formatWithEscapedParameter(String format, String parameter, Long index)
   {
      if (parameter.startsWith("`"))
      {
         parameter = parameter.substring(1, parameter.length() - 1);
         format = '`' + format + '`';
      }
      if (index == null)
      {
         return String.format(format, parameter);
      }
      else
      {
         return String.format(format, parameter, index);
      }
   }

   /**
    * Returns prefixed parameter optionally containing backticks.
    *
    * @param    parameter
    *           Parameter optionally containing backticks.
    * @param    prefix
    *           Prefix.
    *
    * @return   Returns prefixed parameter optionally containing backticks.
    */
   public static String getPrefixedParameter(String parameter, String prefix)
   {
      StringBuilder buf = new StringBuilder();
      if (parameter.startsWith("`"))
      {
         parameter = parameter.substring(1, parameter.length());
         buf.append('`');
      }
      buf.append(prefix);
      buf.append(parameter);
      return buf.toString();
   }

   /**
    * Returns a formatted string using the specified format string and arguments optionally
    * containing backticks.
    *
    * @param    format
    *           Format string.
    * @param    table
    *           String parameter optionally containing backticks.
    * @param    params
    *           String parameters optionally containing backticks.
    *
    * @return   Returns a formatted string using the specified format string and arguments
    *           optionally containing backticks.
    */
   public static String formatWithEscapedParameters(String format,
                                                    String table,
                                                    String... params)
   {
      Object[] formatParams = new Object[params.length + 1];
      if (table.startsWith("`"))
      {
         format = '`' + format + '`';
         formatParams[0] = table.substring(1, table.length() - 1);
         for (int i = 0; i < params.length; i++) {
            if (params[i].startsWith("`"))
            {
               formatParams[i + 1] = params[i].substring(1, params[i].length() - 1);
            }
            else
            {
               formatParams[i + 1] = params[i];
            }
         }
      }
      else
      {
         formatParams[0] = table;
         System.arraycopy(params, 0, formatParams, 1, params.length);
      }
      return String.format(format, formatParams);
   }

   /**
    * Handle an <code>InterruptedException</code>, which may be wrapped in an
    * exception causal chain, by throwing <code>StopConditionException</code>.
    * Alternatively, if the error represents a database connection problem,
    * <code>DatabaseConnectionException</code> is thrown.  Determination of
    * the latter type of problem is dialect-specific.
    * <p>
    * If neither of these conditions is found to be the cause of the error,
    * this method returns normally.
    *
    * @param   database
    *          The database in use when the exception was thrown. May be
    *          <code>null</code> if no database is used.
    * @param   exc
    *          An exception to inspect.
    *
    * @throws  StopConditionException
    *          if <code>exc</code> or any exception in its causal chain is an
    *          instance of <code>InterruptException</code>.
    * @throws  DatabaseConnectionException
    *          if <code>exc</code> or any exception in its causal chain is
    *          determined to be a database connection error.
    */
   public static void handleException(Database database, Exception exc)
   {
      if (database != null)
      {
         P2JDialect dialect = (P2JDialect) DatabaseManager.getDialect(database);
         
         if (dialect.isConnectionError(exc))
         {
            ErrorManager.displayAbend(1006, "Database " + database.getName() + " not connected");
            ErrorManager.displayError(492, "Unable to run startup procedure");
            throw new DatabaseConnectionException(exc, database);
         }
      }
      
      Throwable next = exc;
      while (next != null)
      {
         if (next instanceof InterruptedException)
         {
            throw new StopConditionException(exc);
         }
         next = next.getCause();
      }
   }

   /**
    * Get field reference(s) for ADD-LIKE-COLUMN. Includes error handling.
    *
    * @param  query
    *         The query which backs the browse ADD-LIKE-COLUMN was called for.
    * @param  fieldHandle
    *         Handle to the target field.
    *
    * @return reference to the target field if <code>fieldHandle</code> is valid and
    *         references a buffer which participates in the query, <code>null</code> otherwise.
    *         If <code>fieldHandle</code> references an extent field, multiple field references
    *         for each extent element are returned.
    */
   public static FieldReference[] getFieldForAddLikeColumn(P2JQuery query, handle fieldHandle)
   {
      if (!fieldHandle._isValid())
      {
         return null;
      }

      WrappedResource resource = fieldHandle.getResource();
      if (!BufferField.class.isAssignableFrom(resource.getClass()))
      {
         ErrorManager.recordOrShowError(9162,
               "ADD-LIKE-COLUMN requires valid buffer-field object",
               false, false);
         return null;
      }

      BufferField srcBufferField = (BufferField) resource;
      Buffer srcBuffer = srcBufferField.bufferHandle().unwrapBuffer();

      boolean match = false;
      int bufferCount = query.numBuffers().intValue();
      for (int i = 1; i <= bufferCount; i++)
      {
         Buffer buffer = query.bufferHandle(i).unwrapBuffer();
         if (buffer.equals(srcBuffer))
         {
            match = true;
            break;
         }
      }

      if (!match)
      {
         ErrorManager.recordOrShowError(9163,
               "In ADD-LIKE-COLUMN, Cannot find field in browser's query",
               false, false);
         return null;
      }

      String legacyFieldName = ((Nameable) srcBufferField).name().getValue();
      String propertyName = TableMapper.getPropertyName(srcBuffer, legacyFieldName);
      int extent = srcBufferField.extent().intValue();

      if (extent > 1)
      {
         FieldReference[] res = new FieldReference[extent];
         for (int i = 0; i < extent; i++)
         {
            res[i] = new FieldReference((DataModelObject) srcBuffer, propertyName, i);
         }
         return res;
      }
      else
      {
         return new FieldReference[]{new FieldReference((DataModelObject) srcBuffer,
               propertyName)};
      }
   }

   /**
    * Get field reference(s) for ADD-LIKE-COLUMN. Includes error handling.
    *
    * @param  query
    *         The query which backs the browse ADD-LIKE-COLUMN was called for.
    * @param  fieldExpression
    *         Expression which specifies the target field. Has format <code>[database
    *         name.]{legacy table name|legacy buffer name}.legacy field name</code>.
    *
    * @return reference to the target field if the field expression is valid and references a
    *         field of a buffer which participates in the query, <code>null</code> otherwise.
    *         If the field expression references an extent field, multiple field references
    *         for each extent element are returned.
    */
   public static FieldReference[] getFieldForAddLikeColumn(P2JQuery query,
                                                           character fieldExpression)
   {
      String fullSource = fieldExpression.getValue();
      if (fullSource == null)
      {
         return null;
      }

      int pos = fullSource.lastIndexOf('.');
      if (pos < 0)
      {
         ErrorManager.recordOrShowError(9158,
               String.format("%s in ADD-LIKE-COLUMN must have a table qualifier", fullSource),
               false, false);
         return null;
      }

      String srcFieldName = fullSource.substring(pos + 1);
      String fullTableName = fullSource.substring(0, pos);
      String[] tblName = DynamicTablesHelper.parseTableName(fullTableName);
      String srcBufferName = tblName[0];
      String srcDatabaseName = tblName[1];

      // Buffer selection priorities.
      // 1 - Same buffer name, same database.
      // 2 - Default buffer name (i.e. table name), same database.
      // 3 - Same buffer name, no database specified.
      // 4 - Default buffer name, no database specified.
      // 5 - No match.
      // TODO works differently in GUI!

      Buffer selectedBuffer = null;
      int selectedPriority = 5;
      int bufferCount = query.numBuffers().intValue();
      for (int i = 1; i <= bufferCount; i++)
      {
         Buffer buffer = query.bufferHandle(i).unwrapBuffer();
         RecordBuffer recordBuffer = ((BufferImpl) buffer).buffer();

         String bufferName = buffer.name().getValue();
         String defaultBufferName = TableMapper.getLegacyName(buffer);
         String databaseName = recordBuffer.getDatabase().getName();

         boolean databaseMatch = srcDatabaseName != null && srcDatabaseName.equals(databaseName);
         boolean nameMatch = srcBufferName.equals(bufferName);
         boolean defaultNameMatch = srcBufferName.equals(defaultBufferName);

         int priority = 5;
         if (databaseMatch && nameMatch)
            priority = 1;
         else if (databaseMatch && defaultNameMatch)
            priority = 2;
         else if (nameMatch)
            priority = 3;
         else if (defaultNameMatch)
            priority = 4;

         if (priority < selectedPriority)
         {
            selectedBuffer       = buffer;
            selectedPriority     = priority;
         }
      }

      if (selectedBuffer == null)
      {
         ErrorManager.recordOrShowError(9159,
               String.format("Table %s specified in ADD-LIKE-COLUMN not found in browser's QUERY",
                     fullTableName),
               false, false);
         return null;
      }

      int extentIndex = -1;
      if (srcFieldName.matches("^.*\\[.*]$"))
      {
         pos = srcFieldName.lastIndexOf('[');
         String indexStr = srcFieldName.substring(pos + 1, srcFieldName.length() - 1);
         try
         {
            extentIndex = Integer.parseInt(indexStr);
         }
         catch (NumberFormatException e)
         {
            // suppress
         }

         if (extentIndex == -1)
         {
            ErrorManager.recordOrShowError(9160,
                  String.format("In ADD-LIKE-COLUMN, invalid array index specified for %s",
                        srcFieldName),
                  false, false);
            return null;
         }

         srcFieldName = srcFieldName.substring(0, pos);
      }

      String property = TableMapper.getPropertyName(selectedBuffer, srcFieldName);
      RecordBuffer recordBuffer = RecordBuffer.get((DataModelObject) selectedBuffer);
      int extentSize = (int) TempTableBuilder.getExtentSize(
            recordBuffer.getDMOImplementationClass(),
            recordBuffer.getSizerMap().get(property));
      if (extentSize == 1)
         extentSize = 0; // 4GL logic

      if (property == null)
      {
         ErrorManager.recordOrShowError(9161,
               String.format("In ADD-LIKE-COLUMN, could not find source field %s", srcFieldName),
               false, false);
         return null;
      }

      FieldReference[] res;
      if (extentIndex != -1)
      {
         if (extentIndex < 1 || extentIndex > extentSize)
         {
            ErrorManager.recordOrShowError(7365,
                  String.format("%s BUFFER-VALUE argument must be valid array index, less than " +
                        "or equal to field extent %d", srcFieldName, extentSize),
                  false, false);
            return null;
         }

         res = new FieldReference[] {new FieldReference((DataModelObject) selectedBuffer,
               property,
               extentIndex - 1)};
      }
      else if (extentSize > 1)
      {
         res = new FieldReference[extentSize];
         for (int i = 0; i < extentSize; i++)
            res[i] = new FieldReference((DataModelObject) selectedBuffer, property, i);
      }
      else
      {
         res = new FieldReference[] {new FieldReference((DataModelObject) selectedBuffer,
               property)};
      }

      return res;
   }

   /**
    * This method returns return type or parameter type depending on getter/setter method name
    * prefix.
    *
    * @param   method
    *          Method.
    *
    * @return  Return type or parameter type depending on getter/setter method name prefix.
    */
   static Class<?> getGetterSetterType(Method method)
   {
      String methodName = method.getName();
      if (methodName.startsWith("set"))
      {
         Class<?>[] parameterTypes = method.getParameterTypes();

         // if indexed: first parameter is array index, second one is value
         return parameterTypes[parameterTypes.length - 1];
      }
      else if (methodName.startsWith("get") ||  methodName.startsWith("is"))
      {
         return method.getReturnType();
      }
      else
      {
         throw new UnsupportedOperationException(
            "Method getGetterSetterType is acceptable only for getters/setters");
      }
   }

   /**
    * Look up the getter method associated with the given property.  Tests
    * for both <code>get</code> and <code>is</code> variants.
    *
    * @param   dmoClass
    *          DMO implementation class.
    * @param   propName
    *          Unqualified DMO property name.
    * @param   parmTypes
    *          Array of data types which are expected by the getter method.
    *          This will generally be <code>null</code> (or an empty array),
    *          or an array with one element of type <code>Integer.TYPE</code>
    *          (representing the index value passed to an indexed getter
    *          method).
    *
    * @return  Getter method for the given property.
    *
    * @throws  PersistenceException
    *          if <code>propName</code> does not represent a property of
    *          <code>dmoClass</code>.
    */
   static Method findDMOGetterMethod(Class<?> dmoClass, String propName, Class<?>[] parmTypes)
   throws PersistenceException
   {
      String methName;
      
      try
      {
         // First try with a "get" prefix (the majority of cases).
         methName = StringHelper.generateBeanName(propName, true, false);
         
         return dmoClass.getDeclaredMethod(methName, parmTypes);
      }
      catch (NoSuchMethodException exc)
      {
      }
      
      try
      {
         // As a fallback, try with an "is" prefix, in case property is a logical.
         methName = StringHelper.generateBeanName(propName, true, true);
         
         return dmoClass.getDeclaredMethod(methName, parmTypes);
      }
      catch (NoSuchMethodException exc2)
      {
         throw new PersistenceException(
            dmoClass.getName() + " does not contain property '" + propName + "'");
      }
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name is enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * <p>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is
    * emitted.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO's alias,
    *          interface, and implementation class.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    *
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding the property.
    */
   static void composePropertyName(RecordBuffer buffer, String name, StringBuilder buf)
   throws PersistenceException
   {
      composePropertyName(buffer, name, buf, Database.Type.PRIMARY);
   }
   
   /**
    * Compose a DMO property name within the given string buffer. For
    * non-text properties, this will simply emit the property name, qualified
    * by the proper DMO alias.  For text properties, however, the qualified
    * property name may be enclosed within a SQL function which trims trailing
    * whitespace.  For case-insensitive text properties, this expression is
    * enclosed by a SQL function which uppercases the trimmed result string.
    * For a text property whose qualified name is <code>foo.bar</code>, the
    * end result would be:
    * <pre>
    *    upper(rtrim(foo.bar, ' \t\n\r'))
    * </pre>
    * <p>
    * In cases where a special computed column name is required instead of the
    * enclosing function syntax noted above, such a name is generated instead.
    * <p>
    * For a non-text property <code>foo.bar</code>, <code>foo.bar</code> is
    * emitted.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO's alias,
    *          interface, and implementation class.
    * @param   name
    *          Unqualified DMO property name.
    * @param   buf
    *          String buffer into which results are emitted.
    * @param   type
    *          Database type, which drives the dialect to use to compose the name.
    * 
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding the property.
    */
   static void composePropertyName(RecordBuffer buffer,
                                   String name,
                                   StringBuilder buf,
                                   Database.Type type)
   throws PersistenceException
   {
      boolean isCharacter = false;
      boolean ignoreCase = false;
      String computedColumnPrefix = null;
      
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      if (character.class.equals(getDMOPropertyType(dmoClass, name)))
      {
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         isCharacter = true;
         ignoreCase = !DMOIndex.isCaseSensitive(schema, dmoIface, name);
      }
      
      P2JDialect dialect = null;
      if (isCharacter)
      {
         if (!Database.Type.PRIMARY.equals(type))
         {
            Database secDB = new Database(buffer.getDatabase(), type);
            dialect = (P2JDialect) DatabaseManager.getDialect(secDB);
         }
         
         if (dialect == null)
         {
            dialect = buffer.getDialect();
         }
            
         if (dialect.needsComputedColumns())
         {
            Boolean ccIgnoreCase = DatabaseManager.getIgnoreCase(dmoClass, name);
            if (ccIgnoreCase != null)
            {
               computedColumnPrefix = dialect.getComputedColumnPrefix(!ignoreCase);
            }
         }
      }
      
      composePropertyName(isCharacter,
                          ignoreCase,
                          computedColumnPrefix,
                          buffer.getDMOAlias(),
                          name,
                          buf, 
                          dialect);
   }
}
